import os
from pathlib import Path

from dagster_databricks import databricks_pyspark_step_launcher

from .constants import instance_profile

# TODO revise clusters on small data and figure out
# - if graviton can be used (more cost effective)
# - if only ML or also normal distribution (more machine options) can be used
# TODO then add all the necessary cluster configurations here
# TODO optimize databricks cost/resource consumption


# ---------------------------------------------------
# based on https://github.com/dagster-io/hooli-data-eng-pipelines/blob/3b43ceb2ae4bb20acc014ab2d305e238294b11a4/hooli_data_eng/resources/databricks.py
# https://github.com/dagster-io/dagster/issues/18425#issuecomment-1837225243
# we will not use the step launcher - rather pipes
# cluster:
#           new:
#             aws_attributes:
#               ebs_volume_count: 1
#               ebs_volume_size: 32
#             nodes:
#               node_types:
#                 driver_node_type_id: c5.xlarge
#                 node_type_id: c5.xlarge
#             size:
#               autoscale:
#                 max_workers: 3
#                 min_workers: 1
#             spark_version: 10.4.x-scala2.12
def dev_databricks_cluster_config(
    max_workers: int = 3,
):
    return {
        "autoscale": {"min_workers": 1, "max_workers": max_workers},
        "spark_version": "14.2.x-scala2.12",
        "aws_attributes": {
            "first_on_demand": 1,
            "availability": "SPOT_WITH_FALLBACK",
            "zone_id": "us-east-1d",
            "instance_profile_arn": instance_profile,
            "spot_bid_price_percent": 100,
        },
        "node_type_id": "m6id.2xlarge",
        "driver_node_type_id": "md-fleet.xlarge",
        "enable_elastic_disk": True,
    }


db_step_launcher_dev = databricks_pyspark_step_launcher.configured(
    {
        "run_config": {
            "run_name": "launch_step",
            # "cluster": {"new": cluster_config},
            # "cluster": {"existing": {"env": "DATABRICKS_EXISTING_CLUSTER_ID"}},
            "cluster": {"existing": "1201-111856-56hu41et"},
            # "cluster": {"existing": "1201-111856-56hu41et"},
            "libraries": [
                # dagster-databricks==0.21.10
                # PyPI
                # -
                # -
                # dagster-pyspark==0.21.10
                # PyPI
                # -
                # -
                # dagster==1.5.10
                # {"pypi": {"package": f"dagster-aws {dagster_library_version}"}},
                # {"pypi": {"package": f"dagster-cloud {dagster_version}"}},
                {"pypi": {"package": "databricks-sdk<0.66"}},
                # {"pypi": {"package": f"dagster {dagster_version}"}},
                # {"pypi": {"package": f"dagster-pyspark {dagster_library_version}"}},
                # {"pypi": {"package": f"dagster-databricks {dagster_library_version}"}},
                {"pypi": {"package": "fastwarc==0.15.0"}},
                {"pypi": {"package": "surt==0.3.1"}},
                {"pypi": {"package": "warcio==1.7.4"}},
            ],
        },
        "databricks_host": {"env": "DATABRICKS_HOST"},
        "permissions": {},
        # TODO is this needed
        "staging_prefix": "/dbfs/tmp",
        # TODO is this needed
        "wait_for_logs": True,
        # TODO is this needed? What for long runnign jobs?
        "max_completion_wait_time_seconds": 3600,
        # TODO what is the right path here
        # TODO how does this work with 2 packages (library and pipeline) should we inline one?
        # TODO dagster docs say something like legacy here - figure out what is the right thing to do
        "local_pipeline_package_path": str(Path(__file__).parent.parent.parent.parent),
        "env_variables": {
            "DAGSTER_CLOUD_IS_BRANCH_DEPLOYMENT": os.getenv(
                "DAGSTER_CLOUD_IS_BRANCH_DEPLOYMENT", ""
            ),
            "DAGSTER_CLOUD_DEPLOYMENT_NAME": os.getenv(
                "DAGSTER_CLOUD_DEPLOYMENT_NAME", ""
            ),
            "ASCII_WANDB": os.getenv("ASCII_WANDB", ""),
        },
        "secrets_to_env_variables": [
            # {"name": "DATABRICKS_HOST", "key": "adls2_key", "scope": "dagster-test"},
        ],
        # should be handled from instance profile
        # "storage": {
        #     "s3": {
        #         "access_key_key": "access_key_key",
        #         "secret_key_key": "secret_key_key",
        #         "secret_scope": "dagster-test",
        #     }
        # },
    }
)



import glob
import os
import subprocess
from pathlib import Path

from ascii_library.orchestration.pipes import ExecutionMode
from ascii_library.utils.determine_env import (
    get_dagster_deployment_environment,
)


def library_to_cloud_paths(lib_name: str, filesystem: str = "s3"):
    dagster_deployment = get_dagster_deployment_environment()
    if filesystem == "dbfs":
        # TODO: potential race condition for parallel runs. fix version number
        return f"{filesystem}:/customlibs/{dagster_deployment}/{lib_name}-0.0.0-py3-none-any.whl"
    elif filesystem == "without":
        return f"customlibs/{dagster_deployment}/{lib_name}"
    else:
        # TODO: should it be elif?
        return f"customlibs/{dagster_deployment}/{lib_name}-0.0.0-py3-none-any.whl"


def library_from_dbfs_paths(dbfs_path: str):
    last_part = dbfs_path.split("/")[-1]
    return last_part.split("-")[0]


def empty_dir(path: Path) -> None:  # noqa: C901
    """
    Ensure `path` exists and is empty.
    Removes files, dirs, and symlinks under it using pathlib only.
    """
    path.mkdir(parents=True, exist_ok=True)

    # Post-order: deepest paths first (so directories are empty before rmdir)
    for p in sorted(
        path.rglob("*"), key=lambda x: x.as_posix().count("/"), reverse=True
    ):
        try:
            if p.is_dir() and not p.is_symlink():
                # _chmod_writable(p)
                p.rmdir()  # only works when empty (we emptied children first)
            else:
                # _chmod_writable(p)
                p.unlink(missing_ok=True)  # files or symlinks
        except FileNotFoundError:
            pass  # race-safe
        except PermissionError:
            # try once more after forcing writable
            # _chmod_writable(p)
            try:
                p.rmdir() if (p.is_dir() and not p.is_symlink()) else p.unlink(
                    missing_ok=True
                )
            except Exception:
                raise  # bubble up if it really won't go


def ensure_empty(pathlike) -> Path:
    p = Path(pathlike).resolve()
    empty_dir(p)
    return p


def package_library(mylib_path):  # noqa
    mylib_path = os.path.abspath(mylib_path)
    dist_path = os.path.join(mylib_path, "dist")
    build_path = os.path.join(mylib_path, "build")
    # Clear the dist directory if it already exists
    if os.path.exists(dist_path):
        for f in glob.glob(os.path.join(dist_path, "*")):
            os.remove(f)
    else:
        os.makedirs(dist_path)
    if os.path.exists(build_path):
        empty_dir(Path(build_path))

    subprocess.check_call(
        ["python", "-m", "build", "--wheel", "--outdir", dist_path], cwd=mylib_path
    )
    wheel_files = glob.glob(os.path.join(dist_path, "*.whl"))
    if wheel_files:
        wheel_path = wheel_files[
            0
        ]  # this assumes that there is only one wheel, maybe we will work with multiple versions
        package_name = os.path.basename(wheel_path)
        return wheel_path, package_name
    else:
        raise FileNotFoundError("No wheel file found in the dist directory.")


def get_input_path(io_nodes, part_seed, part_cc, lang):
    if lang == "all":
        return f"{io_nodes}/seed_nodes={part_seed}/crawl_id={part_cc}/main_language=*"
    else:
        return (
            f"{io_nodes}/seed_nodes={part_seed}/crawl_id={part_cc}/main_language={lang}"
        )


def calculate_parallelism(spark, input_path):
    record_count = spark.sparkContext.textFile(input_path).count()
    spark_max_sensible_paralellism = 90000
    if record_count >= spark_max_sensible_paralellism:
        return spark_max_sensible_paralellism
    else:
        return max(200, int(record_count / 4))


def configure_spark(  # noqa: C901
    spark,
    execution_mode,
    compression_codec,
    default_parallelism,
    shuffle_partitions,
    partitionDiscovery_parallelism,
):
    cfgs = [
        ("spark.sql.parquet.compression.codec", compression_codec),
        ("spark.sql.files.maxPartitionBytes", 50 * 1024 * 1024),
        ("spark.databricks.delta.retentionDurationCheck.enabled", "true"),
        ("spark.databricks.delta.vacuum.parallelDelete.enabled", "true"),
        ("spark.sql.sources.partitionOverwriteMode", "dynamic"),
        ("spark.databricks.delta.schema.autoMerge.enabled", "True"),
        ("spark.databricks.delta.schema.autoMerge.enabledOnWrite", "True"),
    ]
    if execution_mode == ExecutionMode.Full:
        if default_parallelism:
            cfgs.append(("spark.default.parallelism", default_parallelism))
        if shuffle_partitions:
            cfgs.append(("spark.sql.shuffle.partitions", shuffle_partitions))
        if partitionDiscovery_parallelism:
            cfgs.append(
                ("spark.sql.shuffle.partitions", partitionDiscovery_parallelism)
            )

    for item in cfgs:
        spark.conf.set(item[0], item[1])  # type: ignore



from functools import cached_property

import dagster as dg
from dagster_ray import PipesRayJobClient, RayResource  # pyrefly: ignore
from ray.job_submission import JobSubmissionClient  # pyrefly: ignore


class PipesRayJobClientLazyLocalResource(dg.ConfigurableResource):
    """
    A resource that provides a PipesRayJobClient.
    It depends on a Ray cluster resource being available.
    """

    ray_cluster: RayResource

    @cached_property
    def pipes_client(self) -> PipesRayJobClient:
        """
        Lazily initializes and returns a PipesRayJobClient connected to the
        provided Ray cluster.
        """
        try:
            import ray  # pyrefly: ignore

            if ray.is_initialized():  # type: ignore[missing-argument]
                ray_address = ray.get_runtime_context().gcs_address  # type: ignore[missing-argument]
                if ray_address:
                    host = ray_address.split(":")[0]
                    dashboard_address = f"http://{host}:8265"
                else:
                    dashboard_address = "http://127.0.0.1:8265"
            else:
                dashboard_address = "http://127.0.0.1:8265"
        except Exception as e:
            dg.get_dagster_logger().warning(f"Could not get Ray address: {e}")
            dashboard_address = "http://127.0.0.1:8265"

        dg.get_dagster_logger().info(
            f"Connecting PipesRayJobClient to Ray at: {dashboard_address}"
        )
        return PipesRayJobClient(client=JobSubmissionClient(address=dashboard_address))

    def run(self, *args, **kwargs):
        return self.pipes_client.run(*args, **kwargs)



import os
import shutil
from typing import Any, List, Mapping, Optional, Sequence, Union

from dagster import (
    AssetExecutionContext,
    AssetsDefinition,
    Config,
    MaterializeResult,
    PartitionsDefinition,
    PipesSubprocessClient,
    asset,
    get_dagster_logger,
)
from databricks.sdk.service import jobs
from pydantic import Field

from ascii_library.orchestration.pipes import (
    LibraryConfig,
    LibraryKind,
    get_engine_by_value,
)
from ascii_library.orchestration.pipes.databricks import PipesDatabricksEnhancedClient
from ascii_library.orchestration.pipes.emr import PipesEmrEnhancedClient
from ascii_library.orchestration.pipes.instance_config import CloudInstanceConfig
from ascii_library.orchestration.pipes.spark_pipes import Engine, SparkPipesResource
from ascii_library.orchestration.resources.emr_constants import pipeline_bucket
from ascii_library.utils.determine_env import (
    get_dagster_deployment_environment,
)

deployment_env = get_dagster_deployment_environment()

# TODO: remove all the ignores on this script


def get_libs_dict(
    cfg: Optional[List[LibraryConfig]],
) -> List[Mapping[str, Union[str, Mapping]]]:
    libs_dict: List[Mapping[str, Union[str, Mapping]]] = []

    if cfg is not None:
        for lib in cfg:
            if lib.kind == LibraryKind.Pypi:
                package_str = (
                    f"{lib.name_id}{lib.version}" if lib.version else lib.name_id
                )
                libs_dict.append({lib.kind.value: {"package": package_str}})
            elif lib.kind == LibraryKind.Wheel:
                lib_name = f"s3://{pipeline_bucket}/{lib.name_id}"
                libs_dict.append({lib.kind.value: lib_name})

    return libs_dict


def update_spot_bid_price_percent(fleet_config, new_spot_bid_price_percent):  # noqa: C901
    if hasattr(fleet_config, "percentageOfOnDemandPrice"):
        fleet_config.percentageOfOnDemandPrice = new_spot_bid_price_percent
    return fleet_config


def generate_uploaded_script_paths(
    local_input_path: str, prefix: str = "dbfs:/"
) -> str:
    """Retrieve file name from path.
    construct full path as: "dbfs:/external_pipes/<<filename>>.py"
    """
    # TODO potential race condition (file in DBFS), in case of parallel runs
    # this is not yet including versioning. When launching more than 1 instance of a databricks spark job in parallel this file is overwritten

    filename_without_extension = os.path.splitext(os.path.basename(local_input_path))[0]
    new_path = f"{prefix}/{deployment_env}/{filename_without_extension}.py"
    return new_path


def get_engine_from_config(config, spark_pipes_client):
    if "config" in config.keys():
        if (
            "override_default_engine" in config["config"].keys()
            and config["config"]["override_default_engine"]
        ):
            override_engine = get_engine_by_value(
                config["config"]["override_default_engine"]
            )
            return spark_pipes_client.get_spark_pipes_client(
                override_engine
            ), override_engine
        else:
            return spark_pipes_client.get_spark_pipes_client(
                spark_pipes_client.engine
            ), spark_pipes_client.engine
    else:
        return spark_pipes_client.get_spark_pipes_client(
            spark_pipes_client.engine
        ), spark_pipes_client.engine


def spark_pipes_asset_factory(  # noqa: C901
    name: str,
    key_prefix: Sequence[str],
    spark_pipes_client: SparkPipesResource,
    external_script_file: str,
    partitions_def: Optional[PartitionsDefinition] = None,
    cfg=None,  # Optional[Config]
    deps: Optional[Sequence[AssetsDefinition]] = None,
    group_name: Optional[str] = None,
    local_spark_config: Optional[Mapping[str, str]] = None,
    libraries_to_build_and_upload: Sequence[str] = None,  # type: ignore
    databricks_cluster_config: Optional[dict[str, Any]] = None,
    libraries_config: Optional[List[LibraryConfig]] = None,
    emr_additional_libraries: Optional[List[LibraryConfig]] = None,
    dbr_additional_libraries: Optional[List[LibraryConfig]] = None,
    emr_job_config: Optional[dict] = None,
    fleet_filters: Optional[CloudInstanceConfig] = None,
    # TODO: in the future support IO manager perhaps for python reading directly from S3 io_manager_key:Optional[str]: None
    # but maybe we intend to keep the offline sync process
    # TODO: should we use asset KWARGS here instead for flexibility?
):
    """
    Construct dagster-pipes based Spark assets for multiple engines: local pyspark and databricks

    Automatically configure the right pipes clients.
    In the case of Databricks: Ensure dependencies and scripts are present and ready to be executed - automatically build the dependent libraries and upload these to DBFS.
    """

    engine_to_use = spark_pipes_client.engine
    # TODO: auto upload is a convenience feature for now in the future this should be a CI pipeline step with a defined version - this task should only be executed once on ci build and not per each task
    if cfg is None:

        @asset(
            name=name,
            compute_kind=engine_to_use.value,
            deps=deps,
            partitions_def=partitions_def,
            group_name=group_name,
            key_prefix=key_prefix,
        )
        def inner_spark_pipes_asset(
            context: AssetExecutionContext,
        ) -> MaterializeResult:
            client_params = handle_shared_parameters(context, {})
            client, real_engine = get_engine_from_config(
                client_params, spark_pipes_client
            )
            return handle_pipeline_modes(context, client_params, client, real_engine)  # type: ignore

    else:

        @asset(
            name=name,
            compute_kind=engine_to_use.value,
            deps=deps,
            partitions_def=partitions_def,
            group_name=group_name,
            key_prefix=key_prefix,
        )
        def inner_spark_pipes_asset(
            context: AssetExecutionContext,
            config: cfg,  # type: ignore
        ) -> MaterializeResult:
            client_params = handle_shared_parameters(context, config)
            client, real_engine = get_engine_from_config(
                client_params, spark_pipes_client
            )
            # TODO this produces a potential race condition?
            os.environ["SPARK_PIPES_ENGINE"] = real_engine.value
            return handle_pipeline_modes(context, client_params, client, real_engine)  # type: ignore

    def handle_pipeline_modes(
        context: AssetExecutionContext,
        client_params,
        client: Union[
            PipesSubprocessClient,
            PipesDatabricksEnhancedClient,
            PipesEmrEnhancedClient,
        ],
        real_engine,
    ):
        if real_engine == Engine.Local:
            return handle_local(client_params, context, client)  # type: ignore
        elif real_engine == Engine.Databricks:
            return handle_databricks(client_params, context, client)  # type: ignore
        elif real_engine == Engine.EMR:
            return handle_emr(client_params, context, fleet_filters, client)  # type: ignore
        else:
            raise ValueError(f"Unsupported engine mode: {real_engine.value}")

    def handle_emr(
        client_params, context, fleet_filters, client: PipesEmrEnhancedClient
    ):
        s3_script_path = generate_uploaded_script_paths(
            local_input_path=external_script_file, prefix="external_pipes"
        )
        step_config = {
            "Name": "Spark Step",
            "ActionOnFailure": "TERMINATE_JOB_FLOW",
            "HadoopJarStep": {
                "Jar": "command-runner.jar",
                "Args": ["spark-submit", f"s3://{pipeline_bucket}/{s3_script_path}"],
            },
        }

        emr_job_config["Name"] = client_params["job_name"]  # type: ignore
        if emr_additional_libraries is not None:
            engine_specific_libs = libraries_config.copy()  # type: ignore
            engine_specific_libs.extend(emr_additional_libraries)
        else:
            engine_specific_libs = libraries_config
        if (
            "config" in client_params.keys()
            and "spot_bid_price_percent" in client_params["config"].keys()
        ):
            if fleet_filters is not None:
                fleet_filters = update_spot_bid_price_percent(
                    fleet_filters, client_params["config"]["spot_bid_price_percent"]
                )
        return client.run(  # type: ignore
            context=context,
            emr_job_config=emr_job_config,  # type: ignore
            bucket=pipeline_bucket,
            local_file_path=external_script_file,
            s3_path=s3_script_path,
            step_config=step_config,
            libraries_to_build_and_upload=libraries_to_build_and_upload,  # type: ignore
            libraries=engine_specific_libs,
            extras=client_params,
            fleet_config=fleet_filters,
        ).get_results()

    def handle_databricks(
        client_params, context, client: PipesDatabricksEnhancedClient
    ):
        # we are using databricks engine
        script_file_path_after_upload = generate_uploaded_script_paths(
            external_script_file, prefix="dbfs:/external_pipes"
        )
        if dbr_additional_libraries is not None:
            engine_specific_libs = libraries_config.copy()  # type: ignore
            engine_specific_libs.extend(dbr_additional_libraries)  # type: ignore
        else:
            engine_specific_libs = libraries_config
        if (
            "config" in client_params.keys()
            and "spot_bid_price_percent" in client_params["config"].keys()
        ):
            databricks_cluster_config["spot_bid_price_percent"] = client_params[  # type: ignore
                "config"
            ]["spot_bid_price_percent"]
            databricks_cluster_config["cluster_log_conf"] = {  # type: ignore
                "dbfs": {"destination": "dbfs:/cluster-logs/dagster"}
            }  # type: ignore
            get_dagster_logger().debug(databricks_cluster_config)
        task = jobs.SubmitTask.from_dict(
            {
                "new_cluster": databricks_cluster_config,
                "libraries": get_libs_dict(engine_specific_libs),
                "task_key": "dagster-launched",
                "spark_python_task": {
                    "python_file": script_file_path_after_upload,
                    "source": jobs.Source.WORKSPACE,
                },
            }
        )
        return client.run(  # type: ignore
            task=task,
            context=context,
            env={
                "SPARK_PIPES_ENGINE": "databricks",
            },
            extras=client_params,
            libraries_to_build_and_upload=libraries_to_build_and_upload,  # type: ignore
            local_file_path=external_script_file,
            dbfs_path=script_file_path_after_upload,
        ).get_results()

    def handle_local(client_params, context, client: PipesSubprocessClient):
        cmd = [shutil.which("python"), external_script_file]
        client_params["local_spark_config"] = local_spark_config
        return client.run(  # type: ignore
            command=cmd,  # pyrefly: ignore
            context=context,
            extras=client_params,
        ).get_results()

    def handle_shared_parameters(context, cfg):
        client_params = {
            "execution_mode": spark_pipes_client.execution_mode.value,
            "engine": engine_to_use.value,
            "config": dict(cfg),  # type: ignore
        }
        if partitions_def is not None:
            client_params["partition_key"] = context.partition_key
            job_name = f"{name}_{deployment_env}_{spark_pipes_client.execution_mode.value}_{context.partition_key}"
        else:
            client_params["partition_key"] = None  # pyrefly: ignore
            job_name = (
                f"{name}_{spark_pipes_client.execution_mode.value}_{deployment_env}"
            )
        client_params["job_name"] = job_name
        return client_params

    return inner_spark_pipes_asset


class BaseConfig(Config):
    """Runtime knobs for Spark pipes.

    Fields:
      - ``spot_bid_price_percent``: percent of on-demand price to pay for spot (1–100).
      - ``override_default_engine``: override engine. One of ``pyspark``, ``emr``, ``databricks``.
    """

    spot_bid_price_percent: Optional[int] = Field(  # pyrefly: ignore
        default=90,
        description="Percent of on-demand price to pay (1–100).",
        gt=1,
        le=100,
    )
    override_default_engine: Optional[str] = Field(
        default=None,
        description="Override engine: 'pyspark', 'emr', or 'databricks'.",
    )



from pathlib import Path

import dagster as dg

from ascii_library.sensors.notifications_slack import make_slack_on_failure
from ascii_library.utils.determine_env import get_dagster_deployment_environment


def load_assets(
    prefix: str,
    root_dir: str,
    git_url: str = "https://github.com/ascii-supply-networks/research-space.git",
    git_branch: str = "main",
    make_error_sensor: bool = True,
    ascii_dagster_deployment_url: str = "https://orchestration.ascii.ac.at",
):
    """Load assets from the ascii_library definitions folder.

    Args:
        root_dir (str): The root directory of the respective dagster codelocation project.
    """
    base = dg.load_from_defs_folder(path_within_project=Path(root_dir).parent.parent)

    materializable = [a for a in base.assets if isinstance(a, dg.AssetsDefinition)]  # type: ignore
    passthrough = [
        a
        for a in base.assets  # type: ignore
        if not isinstance(a, dg.AssetsDefinition)
    ]  # SourceAsset

    with_refs = dg.with_source_code_references(materializable)

    linked = dg.link_code_references_to_git(
        assets_defs=with_refs,
        git_url=git_url,
        git_branch=git_branch,
        file_path_mapping=dg.AnchorBasedFilePathMapping(
            local_file_anchor=Path(root_dir).parent,
            file_anchor_path_in_repository=prefix,
        ),
    )

    if make_error_sensor:
        deployment_name = get_dagster_deployment_environment()
        if deployment_name in ["PROD"]:
            slack_failure_sensor = make_slack_on_failure(
                base_url=ascii_dagster_deployment_url
            )
            base.sensors.append(slack_failure_sensor)  # pyrefly: ignore
    return base, linked, passthrough



instance_profile = "arn:aws:iam::018967853140:instance-profile/data-access-profile"

s3_log_uri = "s3n://aws-logs-018967853140-us-east-1/elasticmapreduce/"

master_security_group = "sg-0f9b7e8c1d91cb290"
slave_security_group = "sg-0d69b38ec2d7da5db"
service_role = "service-role/AmazonEMR-ServiceRole-20240117T104113"
job_role = "arn:aws:iam::018967853140:instance-profile/AmazonEMR-InstanceProfile-20240117T104057"
aws_region = "us-east-1"
rackspace_user = "018967853140"



from dagster import DagsterError


class CustomPipesException(DagsterError):
    """Custom exception for handling errors with cloud job flow."""

    def __init__(self, message):
        super().__init__(message)



from dataclasses import dataclass
from enum import Enum
from typing import Optional


class LibraryKind(Enum):
    Pypi = "pypi"
    Wheel = "whl"


@dataclass
class LibraryConfig:
    """Configuration for a Python library dependency.

    - For pypi: The PyPI library name and version.
    - For whl: The file path to the library (e.g., s3://... or dbfs:/...).

    Version specifiers should be included (e.g., `>=1.1.1` for a range or `==2.0.0` for a specific version).
    """

    kind: LibraryKind
    name_id: str
    version: Optional[str] = None
    extra_flags: Optional[str] = None


class Engine(Enum):
    Local = "pyspark"
    Databricks = "databricks"
    EMR = "emr"


def get_engine_by_value(value: str) -> Engine:
    for engine in Engine:
        if engine.value == value:
            return engine
    raise ValueError(f"No matching Engine for value: {value}")


class ExecutionMode(Enum):
    Full = "full"
    SmallDevSampleS3 = "small_dev_sample_s3"
    # local mode MUST be paired with Engine.Local (pyspark local)
    SmallDevSampleLocal = "small_dev_sample_local"



import re

import pandas as pd
import yaml

try:
    import pyspark
except ImportError:
    print("you may need pyspark for the full functionality of this package")


def clean_column_name(column_name: str) -> str:
    """Clean up a column name by converting it to snake case."""
    # https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
    column_name = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", column_name)
    column_name = re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", column_name)
    column_name = column_name.strip()
    column_name = re.sub(r"\s+", " ", column_name)
    column_name = (
        column_name.lower()
        .strip()
        .replace(" ", "_")
        .replace(".", "_")
        .replace(":", "_")
        .replace("-", "_")
        .replace(",_", "_")
        .replace("__+", "_")
        .replace("(", "")
        .replace(")", "")
        .replace("ä", "ae")
        .replace("ü", "ue")
        .replace("ö", "oe")
        .replace("ß", "ss")
        .replace("&", "_")
        .replace("/", "_per_")
        .replace("\n", "_")
        .replace("%", "_percent")
        .replace("=", "_eq_")
    )

    column_name = re.sub(r"__+", "_", column_name)
    column_name = column_name.strip("_")
    return column_name


def clean_pandas_columns(df: pd.DataFrame):
    """Column cleanup function. Snake cases everything"""
    df.columns = [clean_column_name(col) for col in df.columns]
    return df


def clean_spark_columns(df: pyspark.sql.DataFrame) -> pyspark.sql.DataFrame:  # type: ignore
    """Apply column name cleanup for PySpark DataFrame."""
    for col in df.columns:
        df = df.withColumnRenamed(col, clean_column_name(col))
    return df


# Function to read and modify DBT schema YAML
def clean_dbt_schema_yaml(file_path: str):  # noqa: C901
    """Read a DBT schema YAML file, clean column names, and return the modified schema."""
    with open(file_path, "r") as file:
        schema = yaml.safe_load(file)

    # Assuming the schema follows the structure of DBT schema.yml
    if "models" in schema:  # pyrefly: ignore
        for model in schema["models"]:  # pyrefly: ignore
            if "columns" in model:
                for column in model["columns"]:
                    if "name" in column:
                        column["name"] = clean_column_name(column["name"])

    return schema
